from __future__ import annotations

from typing import *


class Key(str):
    def __new__(cls, key_name, valid_type: Type = None, can_be_none: bool = False):
        f"""
        Creates a key that is based on a str-object. 
        
        :param key_name: The name of the key. Note that the name must be compatible as a variable name in Python, meaning no special characters are allowed.   
        :param valid_type: The type to validate the value of the object mapped to this key. E.g., if {valid_type} is {str}, then you cannot map an {int} to the key.
         If {None}, no validation will be done.
        :param can_be_none: Used to tell varvault that this key may be mapped to a value that is set to {None}.
        """
        obj = super().__new__(cls, key_name)
        obj.key_name = key_name
        obj.can_be_none = can_be_none
        obj.valid_type = valid_type
        return obj

    def __eq__(self, other: Key or str):
        if not isinstance(other, (Key, str)):
            return False

        if isinstance(other, Key):
            return self.key_name == other.key_name
        else:
            return self.key_name == other

    def __hash__(self):
        return hash(self.key_name)

    def type_is_valid(self, obj):
        if not self.valid_type:
            # No valid types defined. Return True
            return True
        else:
            if self.can_be_none:
                if obj is None:
                    return True
                if isinstance(obj, self.valid_type):
                    return True
                try:
                    if issubclass(obj, self.valid_type):
                        return True
                except:
                    pass
                return False
            else:
                if isinstance(obj, self.valid_type):
                    return True
                try:
                    if issubclass(obj, self.valid_type):
                        return True
                except:
                    pass
                return False


class Keyring(object):
    """Base class for keys to be used for a vault. A class which extends
    this class must be used for defining your own keyring."""

    @classmethod
    def get_keys_in_keyring(cls) -> Dict[str, Key]:
        """Returns all keys in the keyring as a dict on this format: Dict[str: Key]"""
        keys = dict()
        for key, value in cls.__dict__.items():
            if not isinstance(value, Key):
                continue
            assert key == value, f"The name of a key in a keyring must match the variable's name; varname: {key}, value: {value}"
            keys[key] = value
        return keys

    @classmethod
    def get_key_by_matching_string(cls, key_str: str) -> Key:
        """Returns a key in the keyring that matches the string passed to the function"""
        try:
            return cls.__dict__[key_str]
        except KeyError:
            raise KeyError(f"Failed to get matching key for string: {key_str}. "
                           f"It doesn't appear to exist in the keyring: {cls.get_keys_in_keyring().values()}")



from typing import *

from .keyring import Key


class MiniVault(dict):
    """Class that acts as a very bare-bones vault to represent variables either coming from the vault via get_multiple, or when returning variables form a vaulted function."""

    def __init__(self, d=None, **kwargs):
        if d is None:
            super(MiniVault, self).__init__(kwargs)
        else:
            super(MiniVault, self).__init__(d)

    @staticmethod
    def build(keys: List[Key] or Tuple[Key], values: List[Any] or Tuple[Any]):
        """Builds a minivault object based on iterables of keys and values. Note that if more values are passed to the function than keys,
        the indices for the values that exceed the length of the keys will be ignored."""
        assert isinstance(keys, (list, tuple)), "'keys' must be a list or a tuple"
        assert isinstance(values, (list, tuple)), "'values' must be a list or a tuple"

        assert len(keys) == len(values), "The length of 'keys' and 'values' must be identical"

        data = dict()
        for i in range(len(keys)):
            key = keys[i]
            value = values[i]
            assert isinstance(key, Key), f"Key {key} is not of correct type {Key}"
            assert key.type_is_valid(value), f"Key '{key}' requires type to be '{key.valid_type}', but type for value is '{type(value)}'. " \
                                             f"Is the order of your keys and values passed to '{MiniVault.build.__name__}' wrong?"
            data[key] = value
        return MiniVault(data)



import json
import asyncio
import hashlib
import logging
from types import *
from typing import *

from .keyring import Keyring, Key
from .minivault import MiniVault
from .vaultstructs import VaultStructBase


def md5hash(fname):
    """Get md5 hash of a file using hashlib"""
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def is_serializable(obj: object, logger: logging.Logger = None):
    f"""Function for testing whether or not an object can be serialized by simply trying to do {json.dumps} on the object. Returns {True} if {obj} can be serialized, otherwise {False}."""

    try:
        json.dumps(obj)
        return True
    except (TypeError, OverflowError) as e:
        if logger:
            logger.debug(f"Failed to serialize object: {e}")
        return False


def concurrently(*args_as_iterable: Union[Sized, Iterable], **input_kwargs):
    """
    Decorator for running a coroutine-function concurrently.

    Example:

    ```
    @concurrently([1, 2, 3, 4, 5], (10, 20, 30, 40), const="2")
    async def run(k, v, const=None):
        print(k, v, const)
        return k * v


    if __name__ == '__main__':
        print(run())
    ```

    Output:

    ```
    1 10 const
    2 20 const
    3 30 const
    4 40 const
    [10, 40, 90, 160]  # This is from printing the result from run()
    ```

    Note that the 5th item in the first list was ignored because the second iterable only included 4 elements. This is due to 'zip'.
    Filling out the tuple with None as the 5th element would make it work.

    :param args_as_iterable: The arguments as iterables to pass to the decorated function. The arguments are combined using the builtin keyword 'zip'.
    :param input_kwargs: Args passed as constants to the function, i.e. these variables will be sent to every call for the decorated function.
    """
    import functools
    import asyncio

    def wrap_outer(func: Union[Coroutine, FunctionType, Callable]):
        assert asyncio.iscoroutinefunction(func), f"Function {func.__module__}.{func.__name__} is not defined as a coroutine; define it as 'async def {func.__name__}(...)'"

        @functools.wraps(func)
        def wrap_inner(*args, **kwargs):
            # Don't use args or kwargs; That's not how this is meant to be used
            assert len(args) == 0 and len(kwargs) == 0, f"You should not pass arguments to this function ({func.__module__}.{func.__name__}) when it's decorated with '{concurrently.__module__}.{concurrently.__name__}'. Arguments should come from the decorator."
            return concurrent_execution(func, *args_as_iterable, **input_kwargs)

        return wrap_inner
    return wrap_outer


def concurrent_execution(target: Union[Coroutine, FunctionType, Callable], *inputs, **kwargs):
    """
    Wraps the asyncio API in Python to make it a bit easier to work with.

    Example usage:
    ```
    async def run(arg1, arg2, const=None):
        assert const == "2"
        print(arg1, arg2)
        return (arg1 + arg2) * int(const)

    values = concurrent_execution(run, [1, 2, 3, 4, 5], [10, 20, 30, 40, 50], const="2")
    print(values)
    ```
    Output will be
    ```
    [22, 44, 66, 88, 110]
    ```

    :param target: A callable defined as coroutine via the keyword "async" that takes an arbitrary amount of arguments
    :param inputs: An arbitrary tuple of arguments as iterables. All iterables will be zipped together like this: zipped = list(zip(*inputs))
    :param kwargs: Kwargs that are treated like constants that will be sent to each call of 'target'. Any object in kwargs will NOT be zipped into the other arguments.
    :return: Whatever target returns, but as a list of what it returned.
    """
    assert asyncio.iscoroutinefunction(target), f"'target' ({target.__module__}.{target.__name__}) is not a coroutine function; define it as 'async def {target.__name__}(...)'"

    async def do(_target, *_inputs, **_kwargs):
        zipped = list(zip(*_inputs))
        _ret = await asyncio.gather(*[asyncio.create_task(_target(*i, **_kwargs)) for i in zipped])
        return _ret

    return asyncio.run(do(target, *inputs, **kwargs))


def create_mini_vault_from_file(filename_from: str, keyring: Type[Keyring], **extra_keys) -> MiniVault:
    """Creates a minivault object from a file by."""
    import json

    assert issubclass(keyring, Keyring)
    vault_file_data = dict()

    vault_file_data = json.load(open(filename_from))
    assert isinstance(vault_file_data, dict)

    # Get the keys from the file as a list.
    keys_from_keyring = keyring.get_keys_in_keyring()
    keys_from_keyring.update(extra_keys)
    return_vault_data = dict()

    async def build(key_in_file: str):
        if key_in_file not in keys_from_keyring:
            return
        key: Key = keys_from_keyring[key_in_file]
        if issubclass(key.valid_type, VaultStructBase):
            return_vault_data[key] = key.valid_type.build_from_vault_key(key_in_file, vault_file_data[key_in_file])
        else:
            if key.can_be_none and vault_file_data[key_in_file] is None:
                return_vault_data[key] = None
            else:
                assert isinstance(vault_file_data[key_in_file], key.valid_type), f"Key type missmatch ({key}; Valid type {key.valid_type}, actual type: {type(vault_file_data[key_in_file])}"
                return_vault_data[key] = vault_file_data[key_in_file]

    concurrent_execution(build, vault_file_data.keys())

    return MiniVault(**return_vault_data)



from __future__ import annotations

import warnings


class VaultFlags(str):
    """Class that represents flags that may be used for a vaulted
    function to tweak the behavior of the vaulted function."""

    _RETURN_VALUES_CANNOT_BE_NONE = "return_values_cannot_be_none"
    _PERMIT_MODIFICATIONS = "permit_modifications"
    _INPUT_KEY_CAN_BE_MISSING = "input_key_can_be_missing"
    _CLEAN_RETURN_KEYS = "clean_return_keys"
    _DEBUG = "debug"
    _SILENT = "silent"
    _LIVE_UPDATE = "live_update"
    _RETURN_TUPLE_IS_SINGLE_ITEM = "return_tuple_is_single_item"
    _SPLIT_RETURN_KEYS = "split_return_keys"
    _FILE_IS_READ_ONLY = "file_is_read_only"
    _DISABLE_LOGGER = "disable_logger"
    _IGNORE_KEYS_NOT_IN_KEYRING = "ignore_keys_not_in_keyring"
    _REMOVE_EXISTING_LOG_FILE = "remove_existing_log_file"

    def __new__(cls, name, *args, **kwargs):
        obj = super().__new__(cls, name)
        obj.name = name
        return obj

    def __eq__(self, other):
        if not isinstance(other, (str, VaultFlags)):
            return False
        elif isinstance(other, VaultFlags):
            return self.name == other.name
        elif isinstance(other, str):
            return self.name == other

    @staticmethod
    def flag_is_set(flag: VaultFlags, *flags: VaultFlags):
        """This is not a flag. This function checks if a flag exists among a bunch of flags"""
        return flag in flags

    @staticmethod
    def return_values_cannot_be_none():
        f"""Flag to set if return values must be something other than {None}. By default, this is fine, but you can enforce return variables to be something other than {None}"""
        return VaultFlags(VaultFlags._RETURN_VALUES_CANNOT_BE_NONE)

    @staticmethod
    def permit_modifications():
        f"""Flag to set if variables may be modified either in the vault itself or for a specific decorated function. 
        By default, varvault doesn't permit modifications to existing keys as this can cause unintended behavior."""
        return VaultFlags(VaultFlags._PERMIT_MODIFICATIONS)

    # FIXME: Remove this in 2.0.0
    @staticmethod
    def input_var_can_be_missing():
        f"""Flag to set if an input variable may be missing in a vault when it is accessed. In this case, the key will be sent to kwargs but it will be mapped to {None}."""
        warnings.warn(f"Using deprecated vault-flag {VaultFlags.input_var_can_be_missing.__name__}. "
                      f"This function is deprecated and will be removed in 2.0.0. Please use {VaultFlags.input_key_can_be_missing.__name__} instead",
                      category=DeprecationWarning,
                      stacklevel=2)
        return VaultFlags(VaultFlags._INPUT_KEY_CAN_BE_MISSING)

    @staticmethod
    def input_key_can_be_missing():
        f"""Flag to set if an input variable may be missing in a vault when it is accessed. In this case, the key will be sent to kwargs but it will be mapped to {None}."""
        return VaultFlags(VaultFlags._INPUT_KEY_CAN_BE_MISSING)

    # FIXME: Remove this in 2.0.0
    @staticmethod
    def clean_return_var_keys():
        f"""Flag to clean return keys in a vault defined for a decorated function. This can be used during a cleanup stage. 
        Varvault will try to map the key to a default value for the valid type, like for example str(), or list(). If it doesn't work, the key will be mapped to {None}."""
        warnings.warn(f"Using deprecated vault-flag {VaultFlags.clean_return_var_keys.__name__}. "
                      f"This function is deprecated and will be removed in 2.0.0. Please use {VaultFlags.clean_return_var_keys.__name__} instead",
                      category=DeprecationWarning,
                      stacklevel=2)
        return VaultFlags(VaultFlags._CLEAN_RETURN_KEYS)

    @staticmethod
    def clean_return_keys():
        f"""Flag to clean return keys in a vault defined for a decorated function. This can be used during a cleanup stage. 
        Varvault will try to map the key to a default value for the valid type, like for example str(), or list(). If it doesn't work, the key will be mapped to {None}."""
        return VaultFlags(VaultFlags._CLEAN_RETURN_KEYS)

    @staticmethod
    def debug():
        f"""Flag to enable debug mode for logger output to the console to help you with debugging. By default, varvault will write debug logs to the logfile, but not the console. 
        By setting this, you'll have a much easier time debugging unintended behavior. Using this and {VaultFlags.silent} in conjunction will cancel each other out and make logging the default."""
        return VaultFlags(VaultFlags._DEBUG)

    @staticmethod
    def silent():
        f"""Flag to enable silent mode for a vault. This will completely remove debug logs being written to the logfile. This can be used to reduce unnecessary
         bloat and make debugging much more easy to do. Using this and {VaultFlags.debug} in conjunction will cancel each other out and make logging the default."""
        return VaultFlags(VaultFlags._SILENT)

    @staticmethod
    def live_update():
        f"""Flag to enable live-update of a vault file. If this is set, the vault will try to update its contents from an existing vault file if the contents of the file
         has changed since last time (this is determined by getting an md5 hash of the contents of the file). The live-update is only performed when the vault is accessed via the decorator."""
        return VaultFlags(VaultFlags._LIVE_UPDATE)

    @staticmethod
    def return_tuple_is_single_item():
        f"""Flag to tell varvault that the return value is a tuple that should be mapped to a single return-key. Varvault cannot tell if 
        a tuple is multiple return values or a single item meant for a single key as this how Python handles multiple return values"""
        return VaultFlags(VaultFlags._RETURN_TUPLE_IS_SINGLE_ITEM)

    @staticmethod
    def split_return_keys():
        f"""Flag to tell varvault that the return keys provided in a MiniVault being returned are split between multiple vaults decorating the same function. 
        By default, any return values from a decorated function must be able to be mapped to the keys defined as return keys. If two vaults are taking return values separately, 
        this wouldn't be possible. Usage of this flag REQUIRES that the return value is a MiniVault-object."""
        return VaultFlags(VaultFlags._SPLIT_RETURN_KEYS)

    @staticmethod
    def file_is_read_only():
        f"""Flag to tell varvault that a vault-file used to create a vault from is read-only."""
        return VaultFlags(VaultFlags._FILE_IS_READ_ONLY)

    @staticmethod
    def disable_logger():
        f"""Flag to tell varvault to disable logger completely and not log anything to a log-file."""
        return VaultFlags(VaultFlags._DISABLE_LOGGER)

    @staticmethod
    def ignore_keys_not_in_keyring():
        f"""Flag to ignore keys not in keyring when creating a vault from an existing vault-file. If {VaultFlags.file_is_read_only} is enabled, this will be enabled by default."""
        return VaultFlags(VaultFlags._IGNORE_KEYS_NOT_IN_KEYRING)

    @staticmethod
    def remove_existing_log_file():
        """Flag to tell varvault to delete an existing log file when creating a vault from an existing vault-file"""
        return VaultFlags(VaultFlags._REMOVE_EXISTING_LOG_FILE)

    @staticmethod
    def return_key_can_be_missing():
        """Flag to tell varvault when using a vaulter-decorated function and not returning objects for all keys to not fail and just set the keys defined.
         If this is set, the return variables MUST be inside a MiniVault object, otherwise varvault cannot determine what variable belongs to what key."""



import json
import logging
import os

from typing import *

from . import MiniVault
from .keyring import Keyring
from .vaultflags import VaultFlags
from .vault import VarVault
from .utils import create_mini_vault_from_file


def create_vault(varvault_keyring: Type[Keyring],
                 varvault_vault_name: str,
                 *flags: VaultFlags,
                 varvault_vault_filename_to: str = None,
                 varvault_specific_logger: logging.Logger = None,
                 **extra_keys) -> VarVault:
    """
    Factory-function to help create a Vault-object instead of creating it manually (which is still possible).

    :param varvault_keyring: Describes which keys belong to this Vault.
    :param varvault_vault_name: Used to name the vault and the logger for writing information and debug messages.
    :param varvault_vault_filename_to: Optional filename for a .JSON file to write the arguments in the vault to.
    :param use_logger: Optional argument for telling varvault to use a logger or not. Default behavior is to use a logger object.
    :param flags: Optional argument for defining VaultFlags for this Vault.
     Note that any global VaultFlag will be overridden by VaultFlags defined in a vault-decorator.
    :param varvault_specific_logger: Optional argument for defining your own logger object if you want to use a specific logger rather than varvault's own logger.
    :param extra_keys: Extra keys as a dict to write to. These keys can be defined during runtime, which can sometimes be necessary.
     It is recommended to use pre-determined keys (e.g. constants), but sometimes being more flexible can be useful.
    :return: Vault object based on input to this function.
    """
    if not VaultFlags.flag_is_set(VaultFlags.remove_existing_log_file(), *flags):
        flags = (VaultFlags.remove_existing_log_file(), *flags)
    return VarVault(varvault_keyring, varvault_vault_name, *flags, varvault_vault_filename_from=varvault_vault_filename_to, varvault_vault_filename_to=varvault_vault_filename_to, varvault_specific_logger=varvault_specific_logger, **extra_keys)


def from_vault(varvault_keyring: Type[Keyring],
               varvault_vault_name: str,
               varvault_vault_filename_from: str,
               *flags: VaultFlags,
               varvault_vault_filename_to: str = None,
               varvault_specific_logger: logging.Logger = None,
               **extra_keys) -> VarVault:
    """
    Factory-function to help create a Vault-object from an existing vault file.

    :param varvault_keyring: Describes which keys belong to this Vault.
    :param varvault_vault_name: Used to name the vault and the logger for writing information and debug messages.
    :param varvault_vault_filename_from: Filename for a .JSON file to load variables from.
    :param flags: A set of VaultFlags to tweak the behavior of the vault.
    :param varvault_vault_filename_to: Optional filename to write the variables to. This can be a separate file to the file we read from.
    :param varvault_specific_logger: Optional argument for defining your own logger object if you want to use a specific logger rather than varvault's own logger.
    This can be useful when changing the keyring without it being a real issue.
    :param extra_keys: Extra keys as a dict to write to. These keys can be defined during runtime, which can sometimes be necessary.
    It is recommended to use pre-determined keys (e.g. constants), but sometimes being more flexible can be useful.
    :return: A vault based on vault_filename_from and keyring.
    """
    varvault_vault_filename_from = os.path.expanduser(os.path.expandvars(varvault_vault_filename_from))
    live_update = VaultFlags.flag_is_set(VaultFlags.live_update(), *flags)
    ignore_keys_not_in_keyring = VaultFlags.flag_is_set(VaultFlags.file_is_read_only(), *flags) or VaultFlags.flag_is_set(VaultFlags.ignore_keys_not_in_keyring(), *flags)

    if not os.path.exists(varvault_vault_filename_from) and not live_update:
        raise FileNotFoundError(f"Vault-file {varvault_vault_filename_from} doesn't appear to exist, and you have not requested live-update via the flag {VaultFlags.live_update.__name__}.")

    if varvault_vault_filename_to:
        varvault_vault_filename_to = os.path.expanduser(os.path.expandvars(varvault_vault_filename_to))
    else:
        varvault_vault_filename_to = varvault_vault_filename_from

    def _check_for_keys_not_in_keyring():
        try:
            vault_file_data = json.load(open(varvault_vault_filename_from))
        except FileNotFoundError:
            if VaultFlags.flag_is_set(VaultFlags.live_update(), *flags):
                # No point checking for keys not in the keyring at this point
                return list()
            else:
                raise

        assert isinstance(vault_file_data, dict), f"It appears we were not able to load a JSON from {varvault_vault_filename_from}. Are you sure this is a valid JSON?"
        keys_in_keyring = varvault_keyring.get_keys_in_keyring()
        keys_in_keyring.update(extra_keys)
        keys_in_file = vault_file_data.keys()
        _keys_not_in_keyring = [k for k in keys_in_file if k not in keys_in_keyring]
        assert len(_keys_not_in_keyring) == 0 or ignore_keys_not_in_keyring, \
            f"Keys found in vault-file '{varvault_vault_filename_from}' that are not in the keyring, " \
            f"and you have not set to ignore keys found that are not in the keyring. " \
            f"Keys not in keyring: {_keys_not_in_keyring}"

        return _keys_not_in_keyring
    keys_not_in_keyring = _check_for_keys_not_in_keyring()
    try:
        mini = create_mini_vault_from_file(varvault_vault_filename_from, varvault_keyring, **extra_keys)
    except FileNotFoundError:
        if live_update:
            # This is fine; If live_update is defined, it must be possible for the file to not exist yet.
            mini = MiniVault()
        else:
            raise

    vault = VarVault(varvault_keyring, varvault_vault_name, *flags,
                     varvault_vault_filename_from=varvault_vault_filename_from,
                     varvault_vault_filename_to=varvault_vault_filename_to,
                     varvault_specific_logger=varvault_specific_logger,
                     **extra_keys)
    vault.vault.put(mini)

    if len(keys_not_in_keyring) and ignore_keys_not_in_keyring:
        vault.log(f"Vault was created from file '{varvault_vault_filename_from}', and the file contained keys not in the keyring ({keys_not_in_keyring}), "
                  f"but you asked to ignore if this happens through setting 'ignore_keys_not_in_keyring' to True.", level=logging.WARNING)
    else:
        vault.log(f"Vault created from file '{varvault_vault_filename_from}'.", level=logging.INFO)

    return vault



import abc


class VaultStructBase(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def build_from_vault_key(cls, vault_key, vault_value):
        pass


class VaultStructDictBase(VaultStructBase, dict):
    @classmethod
    @abc.abstractmethod
    def build_from_vault_key(cls, vault_key, vault_value):
        pass

    def __setattr__(self, key, value):
        super(VaultStructDictBase, self).__setattr__(key, value)
        super(VaultStructDictBase, self).__setitem__(key, value)


class VaultStructListBase(VaultStructBase, list):
    @classmethod
    @abc.abstractmethod
    def build_from_vault_key(cls, vault_key, vault_value):
        pass


class VaultStructStringBase(VaultStructBase, str):
    @classmethod
    @abc.abstractmethod
    def build_from_vault_key(cls, vault_key, vault_value):
        pass


class VaultStructIntBase(VaultStructBase, int):
    @classmethod
    @abc.abstractmethod
    def build_from_vault_key(cls, vault_key, vault_value):
        pass


class VaultStructFloatBase(VaultStructBase, float):
    @classmethod
    @abc.abstractmethod
    def build_from_vault_key(cls, vault_key, vault_value):
        pass



from .keyring import Key, Keyring

from .minivault import MiniVault

from .utils import *

from .vault import VarVault

from .vaultflags import VaultFlags

from .vaultfactory import create_vault
from .vaultfactory import from_vault

from .vaultstructs import *


def clear_logs():
    import os
    import glob

    files = glob.glob("/tmp/varvault-logs/*.log")
    for f in files:
        os.remove(f)



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = pathlib.Path(f"{HERE}/README.md").read_text()

# Version handling
MAJOR = 1  # Change this if the previous MAJOR is incompatible with this build. Set MINOR and PATCH to 0
MINOR = 1  # Change this if the functionality has changed, but we are still backwards compatible with previous MINOR versions. Set PATCH to 0
PATCH = 1  # Change this is if we are fixing a bug that doesn't change the functionality. If a bug fix has caused functionality to be changed, see MINOR instead
VERSION = f"{MAJOR}.{MINOR}.{PATCH}"

# This call to setup() does all the work
setup(
    name="varvault",
    version=VERSION,
    description="A package that sets up a key-value vault to store and access variables in a global context.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/data-ductus/varvault",
    author="Calle Holst",
    author_email="calle.holst@dataductus.se",
    license="Apache 2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8"
    ],
    packages=["varvault"],
    include_package_data=True,
)



from __future__ import annotations

import os
import _io
import abc
import json
import asyncio
import logging
import functools
import traceback

from typing import *
from threading import Lock

from .minivault import MiniVault
from .keyring import Keyring, Key
from .utils import concurrent_execution, md5hash, create_mini_vault_from_file, is_serializable
from .vaultflags import VaultFlags
from .logger import get_logger, configure_logger


class VarVaultInterface(abc.ABC):
    # =========================================================================================================================================
    # vaulter
    # =========================================================================================================================================
    @abc.abstractmethod
    def vaulter(self, *flags: VaultFlags, input_keys: Union[Key, list, tuple] = None, return_keys: Union[Key, list, tuple] = None) -> Callable:
        f"""
        Decorator to define a function as a vaulted function.
        A vaulted function works a lot different to a normal function.
        A vault works like a key-value storage. A vaulted function will get its arguments
        by accessing them from the vault, and then store any returned value back in the vault.

        :param input_keys: Can be of type: {Key}, {list}, {tuple}. Default: {None}. Keys must be defined in the
         keyring for this specific vault.
        :param return_keys: Can be of type: {str}, {list}, {tuple}. Default: {None}. Keys must be defined in the
         keyring for this specific vault.
        :param flags: Optional argument for defining some flags for this vaulted function. Flags that have an effect:
         {VaultFlags.debug},
         {VaultFlags.silent},
         {VaultFlags.input_key_can_be_missing},
         {VaultFlags.permit_modifications},
         {VaultFlags.split_return_keys},
         {VaultFlags.return_key_can_be_missing},
         {VaultFlags.clean_return_keys}
        """
        pass

    # ============================================================
    # insert
    # ============================================================
    @abc.abstractmethod
    def insert(self, key: Key, value: object, *flags: VaultFlags):
        f"""
        Inserts a {value} into the vault mapped to {key}.
        
        :param key: The {key} to insert the value to. Type must be {Key} 
        :param value: The object to assign to {key}
        :param flags: An optional set of flags to tweak the behavior of the insert. Flags that have an effect: 
         {VaultFlags.permit_modifications},
         {VaultFlags.return_values_cannot_be_none}, 
         {VaultFlags.debug},
         {VaultFlags.silent}
        """
        pass

    # =======================================================================
    # get_multiple
    # =======================================================================
    @abc.abstractmethod
    def get_multiple(self, keys: List[Key], *flags: VaultFlags) -> MiniVault:
        f"""
        Get multiple objects from the vault that are mapped to keys in {keys}. 
        
        :param keys: A list of keys to get the objects for. Must be a list of {Key} objects
        :param flags: An optional set of flags to tweak the behavior of the get. Flags that have an effect:
         {VaultFlags.debug},
         {VaultFlags.silent},
         {VaultFlags.input_key_can_be_missing}
        :return: A {MiniVault} with all the objects in the vault mapped to the keys, or {None} if it's not in the vault.  
        """
        pass

    # ========================================================
    # get
    # ========================================================
    @abc.abstractmethod
    def get(self, key: Key, *flags: VaultFlags, default=None):
        f"""
        Get an object from the vault that is mapped to {key}. 
        
        :param key: The key to get an object for. Must be of type {Key}
        :param flags: An optional set of flags to tweak the behavior of the get. Flags that have an effect:
         {VaultFlags.debug},
         {VaultFlags.silent},
         {VaultFlags.input_key_can_be_missing}
        :param default: An optional argument to define which value to return as default is the value doesn't exist in the vault. Default is {None}
        :return: The object in the vault mapped to the {key}, or {None} if it's not in the vault.  
        """
        pass


class VarVault(VarVaultInterface):
    class Vault(dict):
        def __init__(self, vault_file: TextIO = None):
            super(VarVault.Vault, self).__init__()
            self.writable_args = dict()
            self.vault_file = vault_file

        def __setitem__(self, key, value):
            data = {key: value}
            if is_serializable(data):
                self.writable_args.update(data)

            super(VarVault.Vault, self).__setitem__(key, value)

        @functools.singledispatchmethod
        def put(self, *args, **kwargs):
            raise NotImplementedError("Not implemented")

        @put.register
        def _mv(self, mini: MiniVault):
            async def _put(item):
                key, value = item
                self.__setitem__(key, value)
            concurrent_execution(_put, mini.items())
            self.write()

        @put.register
        def _k_v(self, key: Key, value: object):
            self.__setitem__(key, value)
            self.write()

        def write(self):
            # Try to write writable_args to vault_file if it has been defined
            if self.vault_file:
                json.dump(self.writable_args, open(self.vault_file.name, "w"), indent=2)

    def __init__(self, varvault_keyring: Type[Keyring], varvault_vault_name: str,
                 *flags: VaultFlags,
                 varvault_vault_filename_from: str = None,
                 varvault_vault_filename_to: str = None,
                 varvault_specific_logger: logging.Logger = None,
                 **extra_keys):
        f"""
        Creates a vault-object. You should ideally create a vault from the existing factory functions. 
        
        :param varvault_keyring: The {Keyring} class used for this vault.  
        :param varvault_vault_name: The name of the vault. This is used when creating a logfile for logging information to. 
        :param flags: A set of flags used to tweak the behavior of the vault object. See {VaultFlags} for what flags can be used and what they do.   
        :param varvault_vault_filename_from: Optional. The name of the file to load a vault from.  
        :param varvault_vault_filename_to: Optional. The name of a JSON file to write data in the vault to.  
        :param varvault_specific_logger: Optional. A specific logger to log to in-case you do not want to use the built-in logger in varvault.  
        :param extra_keys: Optional. A kwargs-object with extra keys that are not defined in the {varvault_keyring}. This can be useful when you have a lot of keys that you might 
         want to handle in a programmatic sense rather than in a pre-defined sense. 
        """
        assert issubclass(varvault_keyring, Keyring), f"'keyring' must be a subclass of {Keyring}"
        self.keyring_class = varvault_keyring
        self.flags: list = list(flags)
        disable_logger = VaultFlags.flag_is_set(VaultFlags.disable_logger(), *self.flags)
        remove_existing_log_file = VaultFlags.flag_is_set(VaultFlags.remove_existing_log_file(), *self.flags)
        self.file_is_read_only = VaultFlags.flag_is_set(VaultFlags.file_is_read_only(), *self.flags)
        self.live_update = VaultFlags.flag_is_set(VaultFlags.live_update(), *self.flags)

        if varvault_specific_logger:
            self.logger = varvault_specific_logger
        else:
            self.logger = get_logger(varvault_vault_name, remove_existing_log_file) if not disable_logger else None

        self.lock = Lock()
        assert not VaultFlags.flag_is_set(VaultFlags.clean_return_keys(), *self.flags), f"You really should not set {VaultFlags.clean_return_var_keys()} " \
                                                                                            f"to the vault itself as that would be an extremely bad idea."

        # Get the keys from the keyring and expand it with extra keys
        self.keys: Dict[str, Key] = self.keyring_class.get_keys_in_keyring()
        for key_name, key in extra_keys.items():
            assert isinstance(key_name, str) and isinstance(key, Key), f"extra_keys is not setup correctly; The pattern {{{key_name} ({type(key_name)}): {key} ({type(key)})}} is incorrect. Correct pattern would be {{{key_name} ({str}): {key} ({Key})}}"
        self.keys.update(extra_keys)

        # Expand vault-files passed to constructor to expand both env vars and user in case they have been defined that way
        varvault_vault_filename_from = os.path.expanduser(os.path.expandvars(varvault_vault_filename_from)) if varvault_vault_filename_from else None
        varvault_vault_filename_to = os.path.expanduser(os.path.expandvars(varvault_vault_filename_to)) if varvault_vault_filename_to else None
        self.vault_filename_from = varvault_vault_filename_from
        self.vault_filename_to = varvault_vault_filename_to

        # Create the vault file object
        self.vault_file = self._create_vault_file() if self.vault_filename_to and not self.file_is_read_only else None

        self.vault_file_from_hash = self._get_vault_file_hash()

        # Create the inner vault
        self.inner_vault = self.Vault(self.vault_file)

        if self.vault_file:
            self.log(f"Vault writing data to '{self.vault_file.name}'", level=logging.INFO)
        if self.live_update:
            self.log(f"Vault doing live updates from '{self.vault_filename_from}' whenever the vault is accessed.", level=logging.INFO)

    def __contains__(self, key: Key):
        self._assert_key_is_correct_type(key, msg=f"{self.__contains__.__name__} may only be used with a {Key}-object, not {type(key)}")

        if key.key_name not in self.keys:
            self.log(f"Key {key} does not exist in the keyring. Trying to check if the vault contains this key will never succeed; Consider removing the call that triggered this warning.", level=logging.WARNING)
            return False

        return key in self.vault

    def __del__(self):
        # Close database_file if it has been defined
        if self.vault_file and isinstance(self.vault_file, _io.TextIOWrapper):
            self.vault_file.close()

    def __str__(self):
        return self.inner_vault.__str__()
    
    def log(self, msg: object, *args, level=logging.DEBUG, exception: BaseException = None):
        assert isinstance(level, int), "Log level must be defined as an integer"
        if self.logger:
            self.logger.log(level, msg, *args, exc_info=exception)

    @property
    def vault(self):
        return self.inner_vault

    def _create_vault_file(self):
        """Creates a database file if one has been defined for this vault"""
        if not self.vault_filename_to:
            return

        dirs_path = os.path.dirname(self.vault_filename_to)
        os.makedirs(dirs_path, exist_ok=True)

        file = open(self.vault_filename_to, "w")
        file.write(json.dumps(dict(), indent=2))
        file.write("\n")
        file.close()
        return file

    # =========================================================================================================================================
    # vaulter
    # =========================================================================================================================================
    def vaulter(self, *flags: VaultFlags, input_keys: Union[Key, list, tuple] = None, return_keys: Union[Key, list, tuple] = None) -> Callable:
        #
        # See abstract method in abstract class for an in-depth docstring.
        #

        input_keys = input_keys if input_keys else list()
        return_keys = return_keys if return_keys else list()

        # Validate the input to the decorator
        self._vaulter__validate_input(input_keys, return_keys)

        # Convert input- and return keys to a list if it's a single key to make it easier to handle
        input_keys, return_keys = self._vaulter__convert_input_keys_and_return_keys(input_keys, return_keys)

        # Assert that the keys are in the keyring
        self._vaulter__assert_keys_in_keyring(input_keys, return_keys)

        def wrap_outer(func):
            func_module_name = f"{func.__module__}.{func.__name__}"

            def pre_call(**kwargs):
                self._try_reload_from_file()
                input_kwargs = self._vaulter__build_input_var_keys(input_keys, kwargs, *flags)
                kwargs.update(input_kwargs)
                all_flags = self._get_all_flags(*flags)
                self._configure_log_levels_based_on_flags(*all_flags)
                assert callable(func)

                self.log(f"======{'=' * len(func_module_name)}=")
                self.log(f">>>>> {func_module_name}:")
                if input_kwargs:
                    self.log(f"-------------")
                    self.log(f"Input kwargs:")
                    for kwarg_key, kwarg_value in input_kwargs.items():
                        self.log(f"{kwarg_key}: ({type(kwarg_value)}) -- {kwarg_value}")
                    self.log(f"-------------")

                input_kwargs.update(kwargs)

                self.log(f"======= Calling {func.__module__}.{func.__name__} ========")
                return input_kwargs

            def post_call(ret):
                self._vaulter__handle_return_vars(ret, return_keys, *flags)
                self.log(f"<<<<< {func.__module__}.{func.__name__}:")
                self.log(f"======{'=' * len(func.__module__ + '.' + func.__name__)}=\n")
                self._reset_log_levels()

            # Separate handling if the decorated function uses the coroutine API
            if asyncio.iscoroutinefunction(func):
                @functools.wraps(func)
                async def wrap_inner_async(*args, **kwargs):
                    #
                    # Do pre-call related stuff
                    #
                    input_kwargs = pre_call(**kwargs)
                    try:
                        ret = await func(*args, **input_kwargs)
                    except Exception as e:
                        self.log(f"Failed to run {func.__module__}.{func.__name__}: {e}", level=logging.ERROR)
                        self.log(str(traceback.format_exc()).rstrip("\n"), level=logging.ERROR)
                        raise

                    #
                    # Do post-call related stuff
                    #
                    post_call(ret)

                    return ret
                return wrap_inner_async
            else:
                @functools.wraps(func)
                def wrap_inner(*args, **kwargs):
                    #
                    # Do pre-call related stuff
                    #
                    input_kwargs = pre_call(**kwargs)
                    try:
                        ret = func(*args, **input_kwargs)
                    except Exception as e:
                        self.log(f"Failed to run {func.__module__}.{func.__name__}: {e}", level=logging.ERROR)
                        self.log(str(traceback.format_exc()).rstrip("\n"), level=logging.ERROR)
                        raise

                    #
                    # Do post-call related stuff
                    #
                    post_call(ret)

                    return ret
                return wrap_inner
        return wrap_outer

    def _vaulter__validate_input(self, input_keys, return_keys, *flags):
        assert isinstance(input_keys, (Key, list, tuple)), f"input_keys must be of type {Key}, {list}, or {tuple}"
        assert isinstance(return_keys, (Key, list, tuple)), f"return_keys must be of type {Key}, {list}, or {tuple}"

        for flag in flags:
            assert isinstance(flag, VaultFlags), f"Flag must be of type {VaultFlags}, not '{type(flag)}'"

    def _vaulter__convert_input_keys_and_return_keys(self, input_keys, return_keys):
        if isinstance(input_keys, Key):
            input_keys = [input_keys]
        if isinstance(return_keys, Key):
            return_keys = [return_keys]

        assert isinstance(input_keys, (list, tuple)), f"Input keys doesn't have the correct type; actual type: {type(input_keys)}"
        assert isinstance(return_keys, (list, tuple)), f"Return keys doesn't have the correct type; actual type: {type(input_keys)}"
        return input_keys, return_keys

    def _vaulter__assert_keys_in_keyring(self, input_keys, return_keys):
        for key in input_keys:
            self._assert_key_is_correct_type(key)
            assert key in self.keys, f"Key '{key}' isn't defined as a key in the keyring"

        for key in return_keys:
            self._assert_key_is_correct_type(key)
            assert key in self.keys, f"Key '{key}' isn't defined as a key in the keyring"

    def _vaulter__build_input_var_keys(self, input_keys, kwargs, *flags):
        mini = self.get_multiple(input_keys, *flags)
        assert len(input_keys) == len(mini)
        for key, value in mini.items():
            assert key not in kwargs, f"Key {key} seems to already exist in kwargs used for the function decorated with '@{self.vaulter.__name__}'"

        return mini

    def _vaulter__handle_return_vars(self, ret, return_keys, *flags):

        if not return_keys:
            # No return keys were defined; Just return from here then as there is nothing else to do.
            return

        all_flags = self._get_all_flags(*flags)

        if VaultFlags.flag_is_set(VaultFlags.split_return_keys(), *all_flags):
            assert isinstance(ret, MiniVault), f"If {VaultFlags.split_return_keys()} is defined, you MUST return values in the form of a {MiniVault} object or we cannot determine which keys go where"
            ret = MiniVault({key: value for key, value in ret.items() if key in return_keys})
        if VaultFlags.flag_is_set(VaultFlags.return_key_can_be_missing(), *all_flags):
            assert isinstance(ret, MiniVault), f"If {VaultFlags.return_key_can_be_missing()} is defined, you MUST return values in the form of a {MiniVault} object or we " \
                                               f"cannot determine which keys should be assigned to the vault and which should be skipped."
        if VaultFlags.flag_is_set(VaultFlags.clean_return_keys(), *all_flags):
            self._clean_return_keys(return_keys)
        else:
            mini = self._to_minivault(return_keys, ret, *flags)

            async def validate_keys_in_mini_vault(key, can_be_missing=False):
                assert key in mini or can_be_missing, f"Key {key} isn't present in MiniVault; keys in mini: {mini.keys()}. " \
                                                      f"You can set the vault-flag {VaultFlags.return_key_can_be_missing()} to skip this validation step"
            concurrent_execution(validate_keys_in_mini_vault, return_keys, can_be_missing=VaultFlags.flag_is_set(VaultFlags.return_key_can_be_missing(), *all_flags))

            async def validate_keys_in_return_keys(key):
                assert key in return_keys, f"Key {key} isn't defined as a return-key; return keys: {return_keys}"
            concurrent_execution(validate_keys_in_return_keys, mini.keys())

            self._insert_minivault(mini, *flags)

    def _clean_return_keys(self, return_keys: Union[List[Key], Tuple[Key]]):
        self.log(f"Cleaning return var keys: {return_keys}")
        for key in return_keys:
            if not key.valid_type:
                temp = None
                self.log(f"Cleaning key {key} by setting it to None (no valid_type defined for {key})")
            else:
                try:
                    temp = key.valid_type()
                    self.log(f"Cleaning key {key} by setting it to '{temp}' (key.valid_type = {key.valid_type})")
                except:
                    temp = None
                    self.log(f"Cleaning key {key} by setting it to '{None}' (valid_type is defined, but no default constructor appears to exist for {key.valid_type})")
            self.vault.put(key, temp)

    def _to_minivault(self, return_keys, ret, *flags) -> MiniVault:
        all_flags = self._get_all_flags(*flags)
        if isinstance(ret, MiniVault):
            mini = ret
        else:
            # It's not a return vault; Let's turn it into one.
            if isinstance(ret, tuple):
                # It's a tuple, which means it's either meant as a single item, or there are multiple return objects
                if len(return_keys) == 1:
                    assert return_keys[0].valid_type == tuple or VaultFlags.flag_is_set(VaultFlags.return_tuple_is_single_item(), *all_flags), \
                        f"You have defined only a single return key, yet you are returning multiple items, yet the valid type for key {return_keys[0]} is not {tuple}, " \
                        f"nor is {VaultFlags.return_tuple_is_single_item()} set."

                    mini = MiniVault.build(return_keys, [ret])
                else:
                    assert len(return_keys) == len(ret), "The number of return variables and the number of return keys must be identical in order to map the keys to the return variables"
                    mini = MiniVault.build(return_keys, ret)
            else:
                # ret is a single item
                assert len(return_keys) == 1, "There appear to be more than 1 return-key defined, but only a single item that is returned"
                mini = MiniVault.build(return_keys, [ret])
        return mini

    # ============================================================
    # insert
    # ============================================================
    def insert(self, key: Key, value: object, *flags: VaultFlags):
        #
        # See abstract method in abstract class for an in-depth docstring.
        #

        # Assert that key has correct type
        self._assert_key_is_correct_type(key)

        self._insert__assert_key_in_keyring(key)

        self._configure_log_levels_based_on_flags(*self._get_all_flags(*flags))

        # Key must be as an iterable, but value doesn't have to be
        mini = self._to_minivault([key], value, *flags)
        self._insert_minivault(mini, *flags)
        self._reset_log_levels()

    def _insert_minivault(self, mini: MiniVault, *flags):
        with self.lock:
            all_flags = self._get_all_flags(*flags)

            if VaultFlags.flag_is_set(VaultFlags.return_values_cannot_be_none(), *all_flags):
                assert all([v for v in mini.values() if v is not None]), f"Some or all values returned are {None}, and {VaultFlags.return_values_cannot_be_none()} " \
                                                                         f"has been defined which means no return value may be {None}"

            if not VaultFlags.flag_is_set(VaultFlags.permit_modifications(), *all_flags):
                keys_already_in_vault = [k for k in mini.keys() if k in self]
                assert len(keys_already_in_vault) == 0, f"Keys {keys_already_in_vault} are already in the vault and {VaultFlags.permit_modifications()} is not set. " \
                                                        f"By default, modifications to existing keys are not permitted."
            self.log("-------------------")
            self.log("Variables going in:")
            for ret_key, ret_value in mini.items():
                self.log(f"{ret_key}: ({type(ret_value)}) -- {ret_value}")
            self.vault.put(mini)
            self.log("-----------------")

    def _insert__assert_key_in_keyring(self, key: Key, skip=False):
        """Assert that the key is in the keyring. This step is skipped if force==True"""
        if skip:
            # Skip this step
            return
        assert key in self.keys

    def _insert__assert_value_may_be_inserted(self, key: Key, value: object, modifications_permitted=False):
        # Validate that key doesn't already exist in the vault, or that modifications_permitted==True
        assert key not in self or modifications_permitted, f"Key {key} already exists in the vault and modifications to existing variables are not permitted."

        # Validate the type of the value to insert into the vault
        assert key.type_is_valid(value), f"Value of type {type(value)} is not permitted; Valid type: {key.valid_type}"

    # =======================================================================
    # get_multiple
    # =======================================================================
    def get_multiple(self, keys: List[Key], *flags: VaultFlags) -> MiniVault:
        #
        # See abstract method in abstract class for an in-depth docstring.
        #

        all_flags = self._get_all_flags(*flags)
        mini = MiniVault()
        for key in keys:
            self._assert_key_is_correct_type(key)
        with self.lock:
            self._try_reload_from_file()

            self._configure_log_levels_based_on_flags(*all_flags)
            if not VaultFlags.flag_is_set(VaultFlags.input_key_can_be_missing(), *all_flags):
                for key in keys:
                    assert key in self, f"Key {key} is not mapped to an object in the vault; it appears to be missing in the vault. " \
                                        f"You can set the flag '{VaultFlags.input_key_can_be_missing()}' to avoid this, in which case the value will be {None}, or make sure a value is mapped to it."
            [mini.update({key: self.vault.get(key)}) for key in keys]
            self._reset_log_levels()
        return mini

    # ========================================================
    # get
    # ========================================================
    def get(self, key: Key, *flags: VaultFlags, default=None):
        #
        # See abstract method in abstract class for an in-depth docstring.
        #

        all_flags = self._get_all_flags(*flags)
        # Assert that key has correct type
        self._assert_key_is_correct_type(key)

        with self.lock:
            self._try_reload_from_file()

            self._configure_log_levels_based_on_flags(*all_flags)

            if not VaultFlags.flag_is_set(VaultFlags.input_key_can_be_missing(), *all_flags):
                assert key in self, f"Key {key} is not mapped to an object in the vault; it appears to be missing in the vault. " \
                                    f"You can set the flag '{VaultFlags.input_key_can_be_missing()}' to avoid this, in which case the value will be {None}, or make sure a value is mapped to it."

            self.log(f"Getting value for key {key} from vault")
            value = self.vault.get(key, default)

            self._reset_log_levels()
            return value

    def _assert_key_is_correct_type(self, key: Key, can_be_str=False, msg=None):
        # Define the error message based on input
        if not msg:
            if can_be_str:
                msg = f"Key {key} is not of required type {Key} or {str}"
            else:
                msg = f"Key {key} is not of required type {Key}"

        if can_be_str:
            assert isinstance(key, (Key, str)), msg
        else:
            assert isinstance(key, Key), msg

    def _try_reload_from_file(self):
        """Can be used to reload from a file if changes has been made to it before"""
        if self.vault_filename_from and self.live_update:
            current_hash = self._get_vault_file_hash()

            # If current hash is an empty string, that means no hash was able to be acquired. In this case, there is no file to load from; Raise FileNotFoundError
            if current_hash == "":
                raise FileNotFoundError(f"Failed to read from file {self.vault_filename_from} to get the current "
                                        f"contents of it to perform live updates. This means you've created a vault "
                                        f"from a file that doesn't exist and set it to perform live updates, but you "
                                        f"tried to read from the vault-file before the vault-file was created. This "
                                        f"is simply not permitted and you will have to look at your workflow to "
                                        f"figure out an appropriate solution for this problem. ")

            if current_hash != self.vault_file_from_hash:
                mini = create_mini_vault_from_file(self.vault_filename_from, self.keyring_class)
                self.vault_file_from_hash = current_hash
                self.vault.put(mini)

    def _get_vault_file_hash(self):
        if not self.vault_filename_from:
            return ""
        else:
            try:
                return md5hash(self.vault_filename_from)
            except FileNotFoundError as e:
                if self.live_update:
                    # This is fine; It just means the file doesn't exist yet
                    return ""
                else:
                    raise e

    def _get_all_flags(self, *flags):
        all_flags = self.flags.copy()
        all_flags.extend(flags)
        return all_flags

    def _configure_log_levels_based_on_flags(self, *all_flags):
        if not self.logger:
            # No logger has been assigned for this vault. Just return then.
            return
        if VaultFlags.flag_is_set(VaultFlags.silent(), *all_flags) and VaultFlags.flag_is_set(VaultFlags.debug(), *all_flags):
            # Use default logging levels; debug and silent cancel each other out
            configure_logger(self.logger)
        elif VaultFlags.flag_is_set(VaultFlags.silent(), *all_flags):
            configure_logger(self.logger, overall_level=logging.INFO)
        elif VaultFlags.flag_is_set(VaultFlags.debug(), *all_flags):
            configure_logger(self.logger, stream_level=logging.DEBUG)

    def _reset_log_levels(self):
        if not self.logger:
            # No logger has been assigned for this vault. Just return then.
            return
        configure_logger(self.logger, overall_level=logging.DEBUG, stream_level=logging.INFO, file_level=logging.DEBUG)
